@@ -1,14 +1,13 @@
 import base64
 import collections
 import itertools
-import json
 import os
 import math
-import shutil
 import struct
-import tempfile
 import pprint # pylint: disable=unused-import
 
+from dataclasses import dataclass
+
 from panda3d.core import * # pylint: disable=wildcard-import
 try:
     from panda3d import bullet
@@ -24,29 +23,24 @@ else:
     CPTA_stdfloat = CPTA_double
     PTA_stdfloat = PTA_double
 
-load_prc_file_data(
-    __file__,
-    'interpolate-frames #t\n'
-)
 
-GltfSettings = collections.namedtuple('GltfSettings', (
-    'physics_engine',
-    'print_scene',
-    'skip_axis_conversion',
-    'no_srgb',
-    'textures',
-    'legacy_materials',
-    'animations',
-))
-GltfSettings.__new__.__defaults__ = (
-    'builtin', # physics engine
-    False, # print_scene
-    False, # skip_axis_conversion
-    False, # do not load textures as sRGB
-    'ref', # reference external textures
-    False, # use PBR materials
-    'embed', # keep animations in the same BAM file
-)
+@dataclass
+class GltfSettings:
+    collision_shapes: str = 'builtin'
+    skip_axis_conversion: bool = False
+    no_srgb: bool = False
+    textures: str = 'ref'
+    legacy_materials: bool = False
+    skip_animations: bool = False
+    flatten_nodes: bool = False
+
+
+def get_extras(gltf_data):
+    extras = gltf_data.get('extras', {})
+    if not isinstance(extras, dict):
+        # weird, but legal; fail silently for now
+        return {}
+    return extras
 
 
 class Converter():
@@ -81,8 +75,8 @@ class Converter():
         'tangent': GeomEnums.C_other,
         'texcoord': GeomEnums.C_texcoord,
         'color': GeomEnums.C_color,
-        'weights': GeomEnums.C_point,
-        'joints': GeomEnums.C_point,
+        'transform_weight': GeomEnums.C_other,
+        'transform_index': GeomEnums.C_index,
     }
     _ATTRIB_NAME_MAP = {
         'position': InternalName.get_vertex().get_name(),
@@ -100,12 +94,15 @@ class Converter():
 
     def __init__(
             self,
-            indir=Filename.from_os_specific(os.getcwd()),
-            outdir=Filename.from_os_specific(os.getcwd()),
-            settings=GltfSettings()
+            filepath,
+            settings=None
     ):
-        self.indir = indir
-        self.outdir = outdir
+        if not isinstance(filepath, Filename):
+            filepath = Filename.from_os_specific(filepath)
+        if settings is None:
+            settings = GltfSettings()
+        self.filepath = filepath
+        self.filedir = Filename(filepath.get_dirname())
         self.settings = settings
         self.cameras = {}
         self.buffers = {}
@@ -131,7 +128,7 @@ class Converter():
         self.background_color = (0, 0, 0)
         self.active_camera = None
 
-    def update(self, gltf_data, writing_bam=False):
+    def update(self, gltf_data):
         #pprint.pprint(gltf_data)
 
         skip_axis_conversion = (
@@ -174,10 +171,6 @@ class Converter():
         for meshid, gltf_mesh in enumerate(gltf_data.get('meshes', [])):
             self.load_mesh(meshid, gltf_mesh, gltf_data)
 
-        # If we support writing bam 6.40, we can safely write out
-        # instanced lights.  If not, we have to copy it.
-        copy_lights = writing_bam and not hasattr(BamWriter, 'root_node')
-
         # Build scenegraphs
         def add_node(root, gltf_scene, nodeid, jvtmap, cvsmap):
             try:
@@ -186,6 +179,7 @@ class Converter():
                 print("Could not find node with index: {}".format(nodeid))
                 return
 
+            scene_extras = get_extras(gltf_scene)
             node_name = gltf_node.get('name', 'node'+str(nodeid))
             if nodeid in self._joint_nodes:
                 # Handle non-joint children of joints, but don't add joints themselves
@@ -228,8 +222,8 @@ class Converter():
             if nodeid in self.skeletons:
                 self.build_character(panda_node, nodeid, jvtmap, cvsmap, gltf_data)
 
-            if 'extras' in gltf_scene and 'hidden_nodes' in gltf_scene['extras']:
-                if nodeid in gltf_scene['extras']['hidden_nodes']:
+            if 'hidden_nodes' in scene_extras:
+                if nodeid in scene_extras['hidden_nodes']:
                     panda_node = panda_node.make_copy()
 
             if 'mesh' in gltf_node:
@@ -270,8 +264,6 @@ class Converter():
                 if has_light_ext:
                     lightid = gltf_node['extensions'][light_ext]['light']
                     light = self.lights[lightid]
-                    if copy_lights:
-                        light = light.make_copy()
                     lnp = np.attach_new_node(light)
                     if self.compose_cs == CS_zup_right:
                         lnp.set_p(lnp.get_p() - 90)
@@ -313,7 +305,7 @@ class Converter():
                             gltf_data['extensions']['BP_physics_engine']['engine'] == 'bullet'
                         )
                     else:
-                        use_bullet = self.settings.physics_engine == 'bullet'
+                        use_bullet = self.settings.collision_shapes == 'bullet'
                     if use_bullet and not HAVE_BULLET:
                         print(
                             'Warning: attempted to export for Bullet, which is unavailable, falling back to builtin'
@@ -342,11 +334,12 @@ class Converter():
                             intangible
                         )
                     if phynode is not None:
-                        np.attach_new_node(phynode)
-            if 'extras' in gltf_node:
-                for key, value in gltf_node['extras'].items():
-                    np.set_tag(key, str(value))
+                        phynp = np.attach_new_node(phynode)
+                        for geomnode in np.find_all_matches('+GeomNode'):
+                            geomnode.reparent_to(phynp)
 
+            for key, value in get_extras(gltf_node).items():
+                np.set_tag(key, str(value))
 
             for child_nodeid in gltf_node.get('children', []):
                 add_node(np, gltf_scene, child_nodeid, jvtmap, cvsmap)
@@ -359,13 +352,14 @@ class Converter():
                     node.hide()
                 for child in node.get_children():
                     visible_recursive(child, visible)
-            if 'extras' in gltf_scene and 'hidden_nodes' in gltf_scene['extras']:
-                if nodeid in gltf_scene['extras']['hidden_nodes']:
-                    #print('Hiding', np)
-                    visible_recursive(np, False)
-                else:
-                    #print('Showing', np)
-                    visible_recursive(np, True)
+
+            hidden_nodes = scene_extras.get('hidden_nodes', [])
+            if nodeid in hidden_nodes:
+                #print('Hiding', np)
+                visible_recursive(np, False)
+            else:
+                #print('Showing', np)
+                visible_recursive(np, True)
 
             # Check if we need to deal with negative scale values
             scale = panda_node.get_transform().get_scale()
@@ -388,12 +382,15 @@ class Converter():
             scene_root = NodePath(ModelRoot(scene_name))
 
             node_list = gltf_scene['nodes']
-            if 'extras' in gltf_scene and 'hidden_nodes' in gltf_scene['extras']:
-                node_list += gltf_scene['extras']['hidden_nodes']
+            hidden_nodes = get_extras(gltf_scene).get('hidden_nodes', [])
+            node_list += hidden_nodes
 
             for nodeid in node_list:
                 add_node(scene_root, gltf_scene, nodeid, {}, {})
 
+            if self.settings.flatten_nodes:
+                scene_root.flatten_medium()
+
             self.scenes[sceneid] = scene_root
 
         # Set the active scene
@@ -402,11 +399,11 @@ class Converter():
             self.active_scene = self.scenes[sceneid]
         if 'scenes' in gltf_data:
             gltf_scene = gltf_data['scenes'][sceneid]
-            if 'extras' in gltf_scene:
-                if 'background_color' in gltf_scene['extras']:
-                    self.background_color = gltf_scene['extras']['background_color']
-                if 'active_camera' in gltf_scene['extras']:
-                    self.active_camera = gltf_scene['extras']['active_camera']
+            scene_extras = get_extras(gltf_scene)
+            if 'background_color' in scene_extras:
+                self.background_color = scene_extras['background_color']
+            if 'active_camera' in scene_extras:
+                self.active_camera = scene_extras['active_camera']
 
     def load_matrix(self, mat):
         lmat = LMatrix4()
@@ -421,11 +418,14 @@ class Converter():
             return
 
         uri = gltf_buffer['uri']
-        if uri.startswith('data:application/octet-stream;base64'):
+        if uri == '_glb_bin' and buffid == 0:
+            buff_data = gltf_buffer['_glb_bin']
+        elif uri.startswith('data:application/octet-stream;base64') or \
+           uri.startswith('data:application/gltf-buffer;base64'):
             buff_data = gltf_buffer['uri'].split(',')[1]
             buff_data = base64.b64decode(buff_data)
         elif uri.endswith('.bin'):
-            buff_fname = os.path.join(self.indir.to_os_specific(), uri)
+            buff_fname = os.path.join(self.filedir.to_os_specific(), uri)
             with open(buff_fname, 'rb') as buff_file:
                 buff_data = buff_file.read(gltf_buffer['byteLength'])
         else:
@@ -484,37 +484,46 @@ class Converter():
             print("Texture '{}' has no source, skipping".format(texid))
             return
 
+        def load_embedded_image(name, ext, data):
+            if not name:
+                name = f'gltf-embedded-{texid}'
+            img_type_registry = PNMFileTypeRegistry.get_global_ptr()
+            img_type = img_type_registry.get_type_from_extension(ext)
+
+            img = PNMImage()
+            img.read(StringStream(data), type=img_type)
+
+            texture = Texture(name)
+            texture.load(img)
+
+            return texture
+
         source = gltf_data['images'][gltf_tex['source']]
         if 'uri' in source:
             uri = source['uri']
-            def write_tex_image(ext):
-                texname = 'tex{}.{}'.format(gltf_tex['source'], ext)
-                texdata = base64.b64decode(uri.split(',')[1])
-                texfname = os.path.join(self.outdir.to_os_specific(), texname)
-                with open(texfname, 'wb') as texfile:
-                    texfile.write(texdata)
-                return texfname
-            if uri.startswith('data:image/png;base64'):
-                uri = write_tex_image('png')
-            elif uri.startswith('data:image/jpeg;base64'):
-                uri = write_tex_image('jpeg')
+            if uri.startswith('data:'):
+                info, b64data = uri.split(',')
+
+                if not (info.startswith('data:image/') and info.endswith(';base64')):
+                    raise RuntimeError(
+                        f'Unknown data URI: {info}'
+                    )
+
+                name = source.get('name', '')
+                ext = info.replace('data:image/', '').replace(';base64', '')
+                data = base64.b64decode(b64data)
+
+                texture = load_embedded_image(name, ext, data)
             else:
-                uri = Filename.fromOsSpecific(uri)
-                if self.settings.textures == 'copy':
-                    src = os.path.join(self.indir.to_os_specific(), uri)
-                    dst = os.path.join(self.outdir.to_os_specific(), uri)
-                    outdir = os.path.dirname(dst)
-                    os.makedirs(outdir, exist_ok=True)
-                    shutil.copy(src, dst)
-            texture = TexturePool.load_texture(uri, 0, False, LoaderOptions())
+                uri = Filename.from_os_specific(uri)
+                fulluri = Filename(self.filedir, uri)
+                texture = TexturePool.load_texture(fulluri, 0, False, LoaderOptions())
+                texture.filename = texture.fullpath = uri
         else:
-            view = self.get_buffer_view(gltf_data, source['bufferView'])
+            name = source.get('name', '')
             ext = source['mimeType'].split('/')[1]
-            img_type = PNMFileTypeRegistry.get_global_ptr().get_type_from_extension(ext)
-            img = PNMImage()
-            img.read(StringStream(view), type=img_type)
-            texture = Texture(source.get('name', ''))
-            texture.load(img)
+            data = self.get_buffer_view(gltf_data, source['bufferView'])
+            texture = load_embedded_image(name, ext, data)
 
         if 'sampler' in gltf_tex:
             gltf_sampler = gltf_data['samplers'][gltf_tex['sampler']]
@@ -583,7 +592,8 @@ class Converter():
             self.mat_mesh_map[matid] = []
 
         pmat = Material(matname)
-        pbr_fallback = {'index': '__pbr-fallback', 'texCoord': 0}
+        base_color_fallback = {'index': '__pbr-fallback', 'texCoord': 0}
+        metallic_roughness_fallback = {'index': '__pbr-fallback', 'texCoord': 0}
         emission_fallback = {'index': '__emission-fallback', 'texCoord': 0}
         normal_fallback = {'index': '__normal-fallback', 'texCoord': 0}
         texinfos = []
@@ -593,15 +603,17 @@ class Converter():
                 pbrsettings = gltf_mat['pbrMetallicRoughness']
 
                 pmat.set_diffuse(LColor(*pbrsettings.get('baseColorFactor', [1.0, 1.0, 1.0, 1.0])))
-                texinfos.append(pbrsettings.get('baseColorTexture', pbr_fallback))
+                texinfos.append(pbrsettings.get('baseColorTexture', base_color_fallback))
                 if texinfos[-1]['index'] in self.textures:
                     self.make_texture_srgb(self.textures[texinfos[-1]['index']])
+                texinfos[-1]['mode'] = TextureStage.M_modulate
 
             texinfos.append(gltf_mat.get('normalTexture', normal_fallback))
             texinfos[-1]['mode'] = TextureStage.M_normal
         else:
-            if 'extensions' in gltf_mat and 'BP_materials_legacy' in gltf_mat['extensions']:
-                matsettings = gltf_mat['extensions']['BP_materials_legacy']['bpLegacy']
+            mat_extensions = gltf_mat.get('extensions', {})
+            if 'BP_materials_legacy' in mat_extensions:
+                matsettings = mat_extensions['BP_materials_legacy']['bpLegacy']
                 pmat.set_shininess(matsettings['shininessFactor'])
                 pmat.set_ambient(LColor(*matsettings['ambientFactor']))
 
@@ -610,6 +622,7 @@ class Converter():
                     texinfos.append(texinfo)
                     if matsettings['diffuseTextureSrgb'] and texinfo['index'] in self.textures:
                         self.make_texture_srgb(self.textures[texinfo['index']])
+                    texinfos[-1]['mode'] = TextureStage.M_modulate
                 else:
                     pmat.set_diffuse(LColor(*matsettings['diffuseFactor']))
 
@@ -618,6 +631,7 @@ class Converter():
                     texinfos.append(texinfo)
                     if matsettings['emissionTextureSrgb'] and texinfo['index'] in self.textures:
                         self.make_texture_srgb(self.textures[texinfo['index']])
+                    texinfos[-1]['mode'] = TextureStage.M_emission
                 else:
                     pmat.set_emission(LColor(*matsettings['emissionFactor']))
 
@@ -632,33 +646,30 @@ class Converter():
                 pbrsettings = gltf_mat['pbrMetallicRoughness']
 
                 pmat.set_base_color(LColor(*pbrsettings.get('baseColorFactor', [1.0, 1.0, 1.0, 1.0])))
-                texinfos.append(pbrsettings.get('baseColorTexture', pbr_fallback))
+                texinfos.append(pbrsettings.get('baseColorTexture', base_color_fallback))
                 if texinfos[-1]['index'] in self.textures:
                     self.make_texture_srgb(self.textures[texinfos[-1]['index']])
 
                 pmat.set_metallic(pbrsettings.get('metallicFactor', 1.0))
                 pmat.set_roughness(pbrsettings.get('roughnessFactor', 1.0))
-                texinfos.append(pbrsettings.get('metallicRoughnessTexture', pbr_fallback))
-                texinfos[-1]['mode'] = TextureStage.M_selector # Unused, but means FFP and auto shader will ignore
+                texinfos.append(pbrsettings.get('metallicRoughnessTexture', metallic_roughness_fallback))
+                texinfos[-1]['mode'] = TextureStage.M_selector
 
             # Normal map
             texinfos.append(gltf_mat.get('normalTexture', normal_fallback))
             texinfos[-1]['mode'] = TextureStage.M_normal
 
             # Emission map
-            factor = gltf_mat.get('emissiveFactor', [0.0, 0.0, 0.0])
-            pmat.set_emission(LColor(*factor, w=0.0))
-            if 'emissiveTexture' in gltf_mat:
-                texinfos.append(gltf_mat['emissiveTexture'])
-                texinfos[-1]['mode'] = TextureStage.M_add
-            else:
-                # The fallback texture is all-white, so set it to "modulate" to
-                # make sure the model isn't affected by it without simplepbr.
-                texinfos.append(emission_fallback)
-                texinfos[-1]['mode'] = TextureStage.M_modulate
+            pmat.set_emission(LColor(*gltf_mat.get('emissiveFactor', [0.0, 0.0, 0.0]), w=0.0))
+            texinfos.append(gltf_mat.get('emissiveTexture', emission_fallback))
+            texinfos[-1]['mode'] = TextureStage.M_emission
             if texinfos[-1]['index'] in self.textures:
                 self.make_texture_srgb(self.textures[texinfos[-1]['index']])
 
+            # Index of refraction
+            ior_ext = mat_extensions.get('KHR_materials_ior', {})
+            pmat.set_refractive_index(ior_ext.get('ior', 1.5))
+
         double_sided = gltf_mat.get('doubleSided', False)
         pmat.set_twoside(double_sided)
 
@@ -790,7 +801,7 @@ class Converter():
         # Check for morph target columns.
         targets = gltf_primitive.get('targets')
         if targets:
-            target_names = gltf_mesh.get('extras', {}).get('targetNames', [])
+            target_names = get_extras(gltf_mesh).get('targetNames', [])
 
             for i, target in enumerate(targets):
                 if i < len(target_names):
@@ -808,11 +819,12 @@ class Converter():
         is_skinned = 'JOINTS_0' in mesh_attribs
         calc_normals = not 'NORMAL' in mesh_attribs
         calc_tangents = not 'TANGENT' in mesh_attribs
+        normalize_weights = False
 
         for buffview, accs in itertools.groupby(accessors, key=lambda x: x['bufferView']):
             buffview = gltf_data['bufferViews'][buffview]
             accs = sorted(accs, key=lambda x: x.get('byteOffset', 0))
-            is_interleaved = len(accs) > 1 and accs[1]['byteOffset'] < buffview['byteStride']
+            is_interleaved = len(accs) > 1 and accs[1].get('byteOffset', 0) < buffview['byteStride']
 
             varray = GeomVertexArrayFormat()
             for acc in accs:
@@ -827,6 +839,7 @@ class Converter():
                 numeric_type = self._COMPONENT_TYPE_MAP[acc['componentType']]
                 numeric_size = self._COMPONENT_SIZE_MAP[acc['componentType']]
                 content = self._ATTRIB_CONTENT_MAP.get(attrib_name, GeomEnums.C_other)
+                size = numeric_size * num_components
 
                 if '_target' in acc:
                     internal_name = InternalName.get_morph(attrib_name, acc['_target'])
@@ -835,6 +848,11 @@ class Converter():
                 # Add this accessor as a column to the current vertex array format
                 varray.add_column(internal_name, num_components, numeric_type, content)
 
+                # Check if the weights table is using float or integer component
+                # Weights normalization will only be performed on float weights.
+                if attrib_parts[0] == 'weights':
+                    normalize_weights = numeric_type == GeomEnums.NT_float32
+
                 if not is_interleaved:
                     # Start a new vertex array format
                     vformat.add_array(varray)
@@ -843,16 +861,19 @@ class Converter():
                         buffview['buffer'],
                         acc.get('byteOffset', 0) + buffview.get('byteOffset', 0),
                         acc['count'],
-                        buffview.get('byteStride', numeric_size * num_components)
+                        size,
+                        buffview.get('byteStride', size)
                     ))
 
             if is_interleaved:
                 vformat.add_array(varray)
+                stride = buffview.get('byteStride', varray.get_stride())
                 data_copies.append((
                     buffview['buffer'],
-                    buffview['byteOffset'],
+                    buffview.get('byteOffset', 0),
                     accs[0]['count'],
-                    buffview.get('byteStride', varray.get_stride())
+                    stride,
+                    stride,
                 ))
 
         # Copy data from buffers
@@ -860,13 +881,22 @@ class Converter():
         vdata = GeomVertexData(geom_node.name, reg_format, GeomEnums.UH_stream)
 
         for array_idx, data_info in enumerate(data_copies):
-            handle = vdata.modify_array(array_idx).modify_handle()
-            handle.unclean_set_num_rows(data_info[2])
+            buffid, start, count, size, stride = data_info
 
-            buff = self.buffers[data_info[0]]
-            start = data_info[1]
-            end = start + data_info[2] * data_info[3]
-            handle.copy_data_from(buff[start:end])
+            handle = vdata.modify_array(array_idx).modify_handle()
+            handle.unclean_set_num_rows(count)
+
+            buff = self.buffers[buffid]
+            end = start + count * stride
+            if stride == size:
+                handle.copy_data_from(buff[start:end])
+            else:
+                src = start
+                dest = 0
+                while src < end:
+                    handle.copy_subdata_from(dest, size, buff[src:src+size])
+                    dest += size
+                    src += stride
             handle = None
 
         # Flip UVs
@@ -894,10 +924,23 @@ class Converter():
                     data = tangent.get_data4f()
                     tangent.set_data4f(data[0], -data[2], data[1], data[3])
 
+        if normalize_weights:
+            # The linear sum of all the joint weights must be as close as possible to 1, if the weights are
+            # stored as float.
+            # Some malformed assets do not respect this, hence we are normalizing them here.
+            weights_data = GeomVertexRewriter(vdata, InternalName.get_transform_weight())
+            while not weights_data.is_at_end():
+                weights = weights_data.get_data4f()
+                max_weight = abs(weights[0]) + abs(weights[1]) + abs(weights[2]) + abs(weights[3])
+                if max_weight != 0.0:
+                    weights = weights / max_weight
+                weights_data.set_data4f(weights)
+
         # Repack mesh data
         vformat = GeomVertexFormat()
         varray_vert = GeomVertexArrayFormat()
         varray_skin = GeomVertexArrayFormat()
+        varray_morph = GeomVertexArrayFormat()
 
         skip_columns = (
             InternalName.get_transform_index(),
@@ -906,9 +949,15 @@ class Converter():
         )
         for arr in reg_format.get_arrays():
             for column in arr.get_columns():
-                varray = varray_skin if column.get_name() in skip_columns else varray_vert
+                column_name = column.get_name()
+                if column_name in skip_columns:
+                    varray = varray_skin
+                elif column_name.parent.basename == "morph":
+                    varray = varray_morph
+                else:
+                    varray = varray_vert
                 varray.add_column(
-                    column.get_name(),
+                    column_name,
                     column.get_num_components(),
                     column.get_numeric_type(),
                     column.get_contents()
@@ -927,6 +976,9 @@ class Converter():
             vformat.add_array(varray_blends)
             vformat.add_array(varray_skin)
 
+        if targets:
+            vformat.add_array(varray_morph)
+
         reg_format = GeomVertexFormat.register_format(vformat)
         vdata = vdata.convert_to(reg_format)
 
@@ -1110,6 +1162,9 @@ class Converter():
     def build_character(self, char, nodeid, jvtmap, cvsmap, gltf_data, recurse=True):
         affected_nodeids = set()
 
+        for bundle in char.bundles:
+            bundle.frame_blend_flag = True
+
         if nodeid in self.skeletons:
             skinid = self.skeletons[nodeid]
             gltf_skin = gltf_data['skins'][skinid]
@@ -1131,7 +1186,7 @@ class Converter():
 
         # Find animations that affect the collected nodes.
         #print("Looking for actions for", skinname, node_ids)
-        if self.settings.animations != 'skip':
+        if not self.settings.skip_animations:
             anims = [
                 (animid, anim)
                 for animid, anim in enumerate(gltf_data.get('animations', []))
@@ -1152,7 +1207,7 @@ class Converter():
             time_acc_id = samplers[0]['input']
             time_acc = gltf_data['accessors'][time_acc_id]
             time_bv = gltf_data['bufferViews'][time_acc['bufferView']]
-            start = time_acc.get('byteOffset', 0) + time_bv['byteOffset']
+            start = time_acc.get('byteOffset', 0) + time_bv.get('byteOffset', 0)
             end = start + time_acc['count'] * 4
             time_data = [
                 struct.unpack_from('<f', self.buffers[time_bv['buffer']], idx)[0]
@@ -1258,19 +1313,20 @@ class Converter():
         skeleton = PartGroup(bundle, "<skeleton>")
         jvtmap = {}
 
-        bind_mats = []
-        ibmacc = gltf_data['accessors'][gltf_skin['inverseBindMatrices']]
-        ibmbv = gltf_data['bufferViews'][ibmacc['bufferView']]
-        start = ibmacc.get('byteOffset', 0) + ibmbv['byteOffset']
-        end = start + ibmacc['count'] * 16 * 4
-        ibmdata = self.buffers[ibmbv['buffer']][start:end]
+        bind_mats = {}
+        if 'inverseBindMatrices' in gltf_skin:
+            ibmacc = gltf_data['accessors'][gltf_skin['inverseBindMatrices']]
+            ibmbv = gltf_data['bufferViews'][ibmacc['bufferView']]
+            start = ibmacc.get('byteOffset', 0) + ibmbv.get('byteOffset', 0)
+            end = start + ibmacc['count'] * 16 * 4
+            ibmdata = self.buffers[ibmbv['buffer']][start:end]
 
-        for i in range(ibmacc['count']):
-            mat = struct.unpack_from('<{}'.format('f'*16), ibmdata, i * 16 * 4)
-            #print('loaded', mat)
-            mat = self.load_matrix(mat)
-            mat.invert_in_place()
-            bind_mats.append(mat)
+            for i in range(ibmacc['count']):
+                mat = struct.unpack_from('<{}'.format('f'*16), ibmdata, i * 16 * 4)
+                #print('loaded', mat)
+                mat = self.load_matrix(mat)
+                mat.invert_in_place()
+                bind_mats[i] = mat
 
         def create_joint(parent, nodeid, transform):
             node = gltf_data['nodes'][nodeid]
@@ -1286,7 +1342,7 @@ class Converter():
             joint_mat = LMatrix4.ident_mat()
             if nodeid in gltf_skin['joints']:
                 joint_index = gltf_skin['joints'].index(nodeid)
-                joint_mat = bind_mats[joint_index]
+                joint_mat = bind_mats.get(joint_index, LMatrix4.ident_mat())
                 self._joint_nodes.add(nodeid)
 
             # glTF uses an absolute bind pose, Panda wants it local
@@ -1322,8 +1378,20 @@ class Converter():
                 meshid = gltf_node['mesh']
                 gltf_mesh = gltf_data['meshes'][meshid]
                 weights = gltf_mesh.get('weights')
-                if weights:
-                    target_names = gltf_mesh.get('extras', {}).get('targetNames', [])
+
+                num_targets = 0
+                for gltf_primitive in gltf_mesh['primitives']:
+                    targets = gltf_primitive.get('targets')
+                    if targets:
+                        num_targets = max(len(targets), num_targets)
+
+                if num_targets > 0:
+                    target_names = get_extras(gltf_mesh).get('targetNames', [])
+                    num_targets = max(len(target_names), num_targets)
+
+                    if not weights:
+                        weights = [0] * num_targets
+
                     if len(target_names) < len(weights):
                         target_names += [str(i) for i in range(len(target_names), len(weights))]
 
@@ -1377,7 +1445,7 @@ class Converter():
 
             buff_view = gltf_data['bufferViews'][acc['bufferView']]
             buff_data = self.buffers[buff_view['buffer']]
-            start = acc.get('byteOffset', 0) + buff_view['byteOffset']
+            start = acc.get('byteOffset', 0) + buff_view.get('byteOffset', 0)
 
             if path == 'rotation':
                 end = start + acc['count'] * 4 * 4
@@ -1508,7 +1576,7 @@ class Converter():
                 acc = accessors[0]
                 buff_view = gltf_data['bufferViews'][acc['bufferView']]
                 buff_data = self.buffers[buff_view['buffer']]
-                start = acc.get('byteOffset', 0) + buff_view['byteOffset']
+                start = acc.get('byteOffset', 0) + buff_view.get('byteOffset', 0)
 
                 end = start + acc['count'] * 4
                 weights = list(CPTAFloat(buff_data[start:end]))
@@ -1536,7 +1604,7 @@ class Converter():
                     # We have too many frames.
                     target_weights = target_weights[:num_frames]
 
-                group.set_table(CPTAFloat(target_weights))
+                group.set_table(CPTA_stdfloat(target_weights))
 
         gltf_node = gltf_data['nodes'][nodeid]
 
@@ -1545,7 +1613,7 @@ class Converter():
             gltf_mesh = gltf_data['meshes'][meshid]
             weights = gltf_mesh.get('weights')
             if weights:
-                target_names = gltf_mesh.get('extras', {}).get('targetNames', [])
+                target_names = get_extras(gltf_mesh).get('targetNames', [])
                 if len(target_names) < len(weights):
                     target_names += [str(i) for i in range(len(target_names), len(weights))]
 
@@ -1716,19 +1784,76 @@ class Converter():
                         'MESH'
                     ))
             if geomnode:
-                verts = []
                 for geom in geomnode.get_geoms():
                     vdata = self.read_vert_data(geom.get_vertex_data(), InternalName.get_vertex())
+                    polygons = []
+                    triangle_map = {}
+
                     for prim in geom.primitives:
                         prim_tmp = prim.decompose()
-                        verts += [
-                            vdata[i].get_xyz() for i in
-                            prim_tmp.get_vertex_list()
-                        ]
 
-                polys = zip(*([iter(verts)] * 3))
-                for poly in polys:
-                    solids.append(CollisionPolygon(*poly))
+                        vertices = prim_tmp.get_vertex_list()
+                        for i in range(0, len(vertices), 3):
+                            pos0 = vdata[vertices[i]].xyz
+                            pos1 = vdata[vertices[i + 1]].xyz
+                            pos2 = vdata[vertices[i + 2]].xyz
+
+                            # Find adjacent triangles lying on the same plane.
+                            normal = (pos2 - pos0).cross(pos1 - pos0)
+                            if not normal.normalize():
+                                # Zero-area triangle.
+                                continue
+
+                            # Quantize the normal.
+                            normal = (int(normal[0] * 0x1000 + 0.5),
+                                      int(normal[1] * 0x1000 + 0.5),
+                                      int(normal[2] * 0x1000 + 0.5))
+
+                            key0 = (normal, pos1, pos0)
+                            key1 = (normal, pos2, pos1)
+                            key2 = (normal, pos0, pos2)
+                            if key0 in triangle_map:
+                                poly, pos3 = triangle_map[key0]
+                                quad = (pos0, pos3, pos1, pos2)
+                                if CollisionPolygon.verify_points(*quad) and \
+                                   not CollisionPolygon(*quad).is_concave():
+                                    poly[:] = quad
+                                    del triangle_map[key0]
+                                    del triangle_map[(normal, pos0, pos3)]
+                                    del triangle_map[(normal, pos3, pos1)]
+                                    continue
+
+                            if key1 in triangle_map:
+                                poly, pos3 = triangle_map[key1]
+                                quad = (pos1, pos3, pos2, pos0)
+                                if CollisionPolygon.verify_points(*quad) and \
+                                   not CollisionPolygon(*quad).is_concave():
+                                    poly[:] = quad
+                                    del triangle_map[key1]
+                                    del triangle_map[(normal, pos1, pos3)]
+                                    del triangle_map[(normal, pos3, pos2)]
+                                    continue
+
+                            if key2 in triangle_map:
+                                poly, pos3 = triangle_map[key2]
+                                quad = (pos2, pos3, pos0, pos1)
+                                if CollisionPolygon.verify_points(*quad) and \
+                                   not CollisionPolygon(*quad).is_concave():
+                                    poly[:] = quad
+                                    del triangle_map[key2]
+                                    del triangle_map[(normal, pos2, pos3)]
+                                    del triangle_map[(normal, pos3, pos0)]
+                                    continue
+
+                            if triangle_map.get((normal, pos0, pos1), (None, None))[1] != pos2:
+                                poly = [pos0, pos1, pos2]
+                                triangle_map[(normal, pos0, pos1)] = (poly, pos2)
+                                triangle_map[(normal, pos1, pos2)] = (poly, pos0)
+                                triangle_map[(normal, pos2, pos0)] = (poly, pos1)
+                                polygons.append(poly)
+
+                    solids.extend(CollisionPolygon(*poly) for poly in polygons)
+
         else:
             print("Unknown collision shape ({}) for object ({})".format(shape_type, node_name))
 
@@ -1741,85 +1866,3 @@ class Converter():
             return phynode
         else:
             print("Could not create collision shape for object ({})".format(node_name))
-
-
-def read_glb_chunk(glb_file):
-    chunk_size, = struct.unpack('<I', glb_file.read(4))
-    chunk_type = glb_file.read(4)
-    chunk_data = glb_file.read(chunk_size)
-    return chunk_type, chunk_data
-
-
-def convert(src, dst, settings=None):
-    if settings is None:
-        settings = GltfSettings()
-
-    if not isinstance(src, Filename):
-        src = Filename.from_os_specific(src)
-
-    if not isinstance(dst, Filename):
-        dst = Filename.from_os_specific(dst)
-
-    indir = Filename(src.get_dirname())
-    outdir = Filename(dst.get_dirname())
-
-    get_model_path().prepend_directory(indir)
-    get_model_path().prepend_directory(outdir)
-
-    converter = Converter(indir=indir, outdir=outdir, settings=settings)
-
-    with open(src, 'rb') as glb_file:
-        if glb_file.read(4) == b'glTF':
-            version, = struct.unpack('<I', glb_file.read(4))
-            if version != 2:
-                raise RuntimeError("Only GLB version 2 is supported, file is version {0}".format(version))
-
-            length, = struct.unpack('<I', glb_file.read(4))
-
-            chunk_type, chunk_data = read_glb_chunk(glb_file)
-            assert chunk_type == b'JSON'
-            gltf_data = json.loads(chunk_data.decode('utf-8'))
-
-            if glb_file.tell() < length:
-                #if read_bytes % 4 != 0:
-                #    glb_file.read((4 - read_bytes) % 4)
-                chunk_type, chunk_data = read_glb_chunk(glb_file)
-                assert chunk_type == b'BIN\000'
-                converter.buffers[0] = chunk_data
-
-            converter.update(gltf_data, writing_bam=True)
-        else:
-            # Re-open as a text file.
-            glb_file.close()
-
-            with open(src) as gltf_file:
-                gltf_data = json.load(gltf_file)
-                converter.update(gltf_data, writing_bam=True)
-
-    if settings.print_scene:
-        converter.active_scene.ls()
-
-    if settings.animations == 'separate':
-        for bundlenode in converter.active_scene.find_all_matches('**/+AnimBundleNode'):
-            anim_name = bundlenode.node().bundle.name
-            anim_dst = dst.get_fullpath_wo_extension() \
-                + f'_{anim_name}.' \
-                + dst.get_extension()
-            bundlenode.write_bam_file(anim_dst)
-    converter.active_scene.write_bam_file(dst)
-
-
-def load_model(loader, file_path, gltf_settings=None, **loader_kwargs):
-    '''Load a glTF file from file_path and return a ModelRoot'''
-
-    with tempfile.NamedTemporaryFile(suffix='.bam') as bamfile:
-        try:
-            bamfilepath = Filename.from_os_specific(bamfile.name)
-            bamfilepath.make_true_case()
-            convert(file_path, bamfilepath, gltf_settings)
-            if hasattr(loader, 'load_sync'):
-                return loader.load_sync(bamfilepath, **loader_kwargs)
-            else:
-                return loader.load_model(bamfilepath, **loader_kwargs)
-        except:
-            raise RuntimeError("Failed to convert glTF file")
