@@ -7,10 +7,9 @@ import panda3d.core as p3d
 
 import simplepbr
 
-import gltf
 
 p3d.load_prc_file_data(
-    __file__,
+    '',
     'window-size 1024 768\n'
     'texture-minfilter mipmap\n'
     'texture-anisotropic-degree 16\n'
@@ -27,8 +26,6 @@ class App(ShowBase):
 
         self.pipeline = simplepbr.init()
 
-        gltf.patch_loader(self.loader)
-
         infile = p3d.Filename.from_os_specific(os.path.abspath(sys.argv[1]))
         p3d.get_model_path().prepend_directory(infile.get_dirname())
 
